/*-----------------------------------------------------------------------*/
/* MMC/SDSC/SDHC (in SPI mode) control module for AT91SAM7               */
/* (C) Martin Thomas, 2009 - based on the AVR module (C)ChaN, 2007       */
/*-----------------------------------------------------------------------*/

/* Copyright (c) 2009, Martin Thomas, ChaN
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.
   * Neither the name of the copyright holders nor the names of
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. */

/*-----------------------------------------------------------------------*/
/*
 * Martin Thomas 9/2009
 * - modified xmit_datablock_dma
 *
 * Martin Thomas 5/2009
 * - based on my older driver from 2006
 * - using ChaN's new driver-skeleton based on his AVR mmc.c
 * - adapted for Olimex SAM7-Pxxx (with AT91SAM7S256)
 *   Hardware SPI, CS @ NCPS0
 * - TODO: disable internal pull-up when external pull-up/downs are installed
 */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/*
 * BestTechnology 3/2011
 * - FULL of PDC
 * - Separate SPI initialization
 */
/*-----------------------------------------------------------------------*/

#include <AT91SAM7S.h>
#include <AT91SAM7S_BITDEFINE.h>
#include <piocfg.h>
#include <stdint.h>
#include "diskio.h"

// Separate SPI initialization
//#define SELF_SPI_INIT

/* Definitions for MMC/SDC command */
#define CMD0    (0x40+0)  /* GO_IDLE_STATE */
#define CMD1    (0x40+1)  /* SEND_OP_COND (MMC) */
#define	ACMD41  (0xC0+41) /* SEND_OP_COND (SDC) */
#define CMD8    (0x40+8)  /* SEND_IF_COND */
#define CMD9    (0x40+9)  /* SEND_CSD */
#define CMD10   (0x40+10) /* SEND_CID */
#define CMD12   (0x40+12) /* STOP_TRANSMISSION */
#define ACMD13  (0xC0+13) /* SD_STATUS (SDC) */
#define CMD16   (0x40+16) /* SET_BLOCKLEN */
#define CMD17   (0x40+17) /* READ_SINGLE_BLOCK */
#define CMD18   (0x40+18) /* READ_MULTIPLE_BLOCK */
#define CMD23   (0x40+23) /* SET_BLOCK_COUNT (MMC) */
#define	ACMD23  (0xC0+23) /* SET_WR_BLK_ERASE_COUNT (SDC) */
#define CMD24   (0x40+24) /* WRITE_BLOCK */
#define CMD25   (0x40+25) /* WRITE_MULTIPLE_BLOCK */
#define CMD55   (0x40+55) /* APP_CMD */
#define CMD58   (0x40+58) /* READ_OCR */

/*--------------------------------------------------------------------------

   DMA/PDC - support ( Martin Thomas )

---------------------------------------------------------------------------*/
/* Dummy-Array to kick SPI-transfers thru PDC for block read.
   If the array is placed in flash (with const) the transfer
   is a little slower but this saves 512 bytes of RAM.
   Define #USE_DUMMY_RAM with 1 to place this array in RAM.
 */
#define DMA_DUMMY_SIZE 512
static const
BYTE dma_dummy[DMA_DUMMY_SIZE] = {
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF };

#ifdef SELF_SPI_INIT
static const TPin pins_on[] = {
  // PA12 -> SPI(MISO)
  {1 << 12,       &PIOA, PIOA_PID, PIO_PERIPH_A, PIO_PULLUP},
  // PA13 -> SPI(MOSI)
  {1 << 13,       &PIOA, PIOA_PID, PIO_PERIPH_A, PIO_DEFAULT},
  // PA14 -> SPI(SPCK)
  {1 << 14,       &PIOA, PIOA_PID, PIO_PERIPH_A, PIO_DEFAULT},
};
#endif

/*--------------------------------------------------------------------------

   Module private portability Functions & Definitions ( Martin Thomas )

---------------------------------------------------------------------------*/

#define SPI_SCBR_MIN          2 /* 1 failed with my cards */
// Chip-Select-Register Number (Array Index of SPI_CSR[])
// here: NPCS0
#define SPI_CSR_NUM           0

#define CARD_SELECT_PIN       (1 << 11)

// select card ( MMC CS = L )
static inline void SELECT (void) {
  PIOA.CODR = CARD_SELECT_PIN;
}

// unselect card MMC CS = H
static inline void DESELECT (void) {
  PIOA.SODR = CARD_SELECT_PIN;
}

#define SOCKWP    0x20      /* Write protect bit-mask (Bit5 set = */
#define SOCKINS   0x10      /* Card detect bit-mask   */

#if 0
/// Pin connected to the Write-Protect WP switch of the card socket
#define SD_SOCKET_WP_PIN      (1 << 16)
/// Pin connected to the Insert/Card Detect switch of the card socket
#define SD_SOCKET_INS_PIN     (1 << 15)
#endif

static inline BYTE get_SOCKWP (void) {
  return 0;
}

static inline BYTE get_SOCKINS (void) {
  return 0;
}

static void init_SOCKWP_SOCKINS (void) {
  return;
}


static inline void set_SOCKPOWER (BOOL on) {
  on = on;
  return;
}

// returns TRUE if power is enabled
static inline BOOL get_SOCKPOWER (void) {
  return TRUE;
}

static void init_SOCKPOWER (void) {
  return;
}

static void AT91_spiSetSpeed (int pcs, BYTE speed) {
  DWORD reg;

  if ( speed < SPI_SCBR_MIN ) speed = SPI_SCBR_MIN;
  if ( speed > 1 ) speed &= 0xFE;
  reg = SPI.CSR[pcs];
  reg = ( reg & ~(_SPI_SCBR) ) | ( (DWORD)speed << 8 );
  SPI.CSR[pcs] = reg;
}

/* general AT91 SPI send/receive */
static BYTE AT91_spi_write_read (BYTE outgoing) {

  void SPI_SendReceive (void *tbuf, void *rbuf, int len) {
    // Check if first bank is free
    while (!(SPI.SR & _SPI_ENDTX)) ;
    if (SPI.PDC.TCR == 0) {
      SPI.PDC.RPR = rbuf;
      SPI.PDC.RCR = len;
      SPI.PDC.TPR = tbuf;
      SPI.PDC.TCR = len;
      SPI.PDC.PTCR = _PDC_TXTEN;
    }
    // Check if second bank is free
    else if (SPI.PDC.TNCR == 0) {
      SPI.PDC.RNPR = rbuf;
      SPI.PDC.RNCR = len;
      SPI.PDC.TNPR = tbuf;
      SPI.PDC.TNCR = len;
    }
    while (!(SPI.SR & _SPI_ENDRX)) ;
  }

  volatile uint32_t tbuf = 0xe0000 | _SPI_LASTXFER | outgoing, rbuf;
  SPI_SendReceive ((void *)&tbuf, (void *)&rbuf, 1);
  return rbuf & 0xff;
}

static inline void rcvr_block_dma (
  BYTE *buff,   /* Data buffer to store received data */
  UINT btr      /* Byte count (must be multiple of 4) */
) {
  volatile DWORD dummy;

  SPI.PDC.RPR = buff;     // destination address
  SPI.PDC.RCR = btr;      // number of frames (here: frame=byte)
  // SPI PDC TX buffer (dummy bytes):
  SPI.PDC.TPR = (void *)dma_dummy; // source address
  SPI.PDC.TCR = btr;      // number of frames (here: frame=byte)
  // enable PDC TX and RX
  SPI.PDC.PTCR = _PDC_TXTEN | _PDC_RXTEN;
  while(!((SPI.SR) & _SPI_RXBUFF)) ; // wait for RX Buffer Full (counters 0)
  dummy = SPI.SR;         // read status to clear flags
}

// seems to work - of it does not work for you select XMIT_TEST 1
// The following may be full of paranoia checks, but a few tests worked as
// expected and I will not change this before more tests can be done.
static inline void xmit_datablock_dma (
  const BYTE *buff  /* 512 byte data block to be transmitted */
) {
  volatile DWORD dummy;

  // SPI PDC TX buffer
  SPI.PDC.RPR   = (void *)&dummy;
  SPI.PDC.RPR   = 0;
  SPI.PDC.RCR   = 0;
  SPI.PDC.RNCR  = 0;
  SPI.PDC.TPR   = (void *)buff; // source address
  SPI.PDC.TCR   = 512;          // number of frames
  // Transmitter transfer enable
  SPI.PDC.PTCR = _PDC_TXTEN;
  while( SPI.PDC.TCR ) ;        // wait for transmit PDC counter 0
  while(!(SPI.SR & _SPI_TXEMPTY)) ; // wait for TDR and shifter empty
  // "When the received data is read, the RDRF bit is cleared."
  dummy = SPI.RDR;
  // "The user has to read the status register to clear the OVRES bit."
  dummy = SPI.SR;
}

/*--------------------------------------------------------------------------

   Module Private Functions ( ChaN )

---------------------------------------------------------------------------*/
static BOOL spi_initalized = FALSE;
static volatile DSTATUS Stat = STA_NOINIT;  /* Disk status */
static volatile BYTE Timer1, Timer2;  /* 100Hz decrement timer */
static BYTE CardType;                 /* b0:MMC, b1:SDv1, b2:SDv2, b3:Block addressing */

/*-----------------------------------------------------------------------*/
/* Init SPI-Interface (Platform dependent)                               */
/*-----------------------------------------------------------------------*/
static void init_spi (void) {
  if (!spi_initalized) {
#ifdef SELF_SPI_INIT
    PIO_Configure (pins_on, PIO_LISTSIZE (pins_on));
    /* set chip-select as output high (unselect card) */
    /* manual control for CS !                        */
    PMC.PCER = ( 1 << PIOA_PID ); // enable needed PIO in PMC
    PIOA.PER  = CARD_SELECT_PIN;  // enable GPIO of CS-pin (disable peripheral functions)
    PIOA.SODR = CARD_SELECT_PIN;  // set high
    PIOA.OER  = CARD_SELECT_PIN;  // output enable

    // enable peripheral clock for SPI ( PID Bit 5 )
    PMC.PCER = ( 1 << SPI_PID ); // n.b. IDs are just bit-numbers
#endif
    SPI.PDC.PTCR = _PDC_RXTDIS | _PDC_TXTDIS;

    // SPI disable
    SPI.CR = _SPI_SPIDIS;
#ifdef SELF_SPI_INIT
    // SPI enable and reset
    // "It seems that the state machine for revB version needs to have 2 SPI
    // software reset to properly reset the state machine."
    SPI.CR = _SPI_SWRST;
    SPI.CR = _SPI_SWRST;

    // SPI mode: master, FDIV=0, fault detection disabled
    SPI.MR = _SPI_MSTR | _SPI_PS_FIXED | _SPI_MODFDIS;// | (0xf << 16);
#endif

    // set chip-select-register
    // 8 bits per transfer, CPOL=1, ClockPhase=0, DLYBCT = 0
    SPI.CSR[SPI_CSR_NUM] = _SPI_NCPHA | _SPI_BITS_8;
    // slow during init
    AT91_spiSetSpeed(SPI_CSR_NUM, 0xFE);

    SPI.PDC.TNPR = SPI.PDC.RNPR = SPI.PDC.TPR = SPI.PDC.RPR = (uint8_t *)0;
    SPI.PDC.TNCR = SPI.PDC.RNCR = SPI.PDC.TCR = SPI.PDC.RCR = 0;
    SPI.PDC.PTCR = _PDC_RXTEN | _PDC_TXTEN;

    // enable SPI
    SPI.CR = _SPI_SPIEN;
    spi_initalized = TRUE;
  }
}

static void close_spi (void) {
  return;
}

/*-----------------------------------------------------------------------*/
/* Transmit a byte to MMC via SPI  (Platform dependent)                  */
/*-----------------------------------------------------------------------*/
static inline void xmit_spi (BYTE dat) {
  AT91_spi_write_read (dat);
}

/*-----------------------------------------------------------------------*/
/* Receive a byte from MMC via SPI  (Platform dependent)                 */
/*-----------------------------------------------------------------------*/
static inline BYTE rcvr_spi (void) {
  return AT91_spi_write_read (0xff);
}

/* replacement for the AVR-Macro */
static inline void rcvr_spi_m (BYTE *dst) {
  *dst = rcvr_spi();
}

/*-----------------------------------------------------------------------*/
/* Wait for card ready                                                   */
/*-----------------------------------------------------------------------*/
static BYTE wait_ready (void) {
  BYTE res;

  Timer2 = 50;	/* Wait for ready in timeout of 500ms */
  rcvr_spi();
  do
    res = rcvr_spi();
  while ((res != 0xFF) && Timer2);

  return res;
}

/*-----------------------------------------------------------------------*/
/* Deselect the card and release SPI bus                                 */
/*-----------------------------------------------------------------------*/
static void release_spi (void) {
  DESELECT();
  rcvr_spi();
}

/*-----------------------------------------------------------------------*/
/* Power Control  (Platform dependent)                                   */
/*-----------------------------------------------------------------------*/
/* When the target system does not support socket power control, there   */
/* is nothing to do in these functions and chk_power always returns 1.   */
static void power_on (void) {
  init_SOCKPOWER ();
  set_SOCKPOWER (TRUE);       /* Socket power ON */
  for (Timer1 = 3; Timer1;) ; /* Wait for 30ms */
  init_SOCKWP_SOCKINS ();     /* Init socket-switches */
  init_spi ();
}

static void power_off (void) {
  SELECT ();                  /* Wait for card ready */
  wait_ready ();
  close_spi ();
  set_SOCKPOWER (FALSE);      /* Socket power OFF */
  Stat |= STA_NOINIT;         /* Set STA_NOINIT */
}

static int chk_power (void) { /* Socket power state: 0=off, 1=on */
  return ( get_SOCKPOWER() ) ? 1 : 0;
}

/*-----------------------------------------------------------------------*/
/* Receive a data packet from MMC                                        */
/*-----------------------------------------------------------------------*/
static BOOL rcvr_datablock (
  BYTE *buff,       /* Data buffer to store received data */
  UINT btr          /* Byte count (must be multiple of 4) */
) {
  BYTE token;

  Timer1 = 10;
  do {              /* Wait for data packet in timeout of 100ms */
    token = rcvr_spi ();
  } while ((token == 0xFF) && Timer1);
  if (token != 0xFE) return FALSE;  /* If not valid data token, return with error */

  rcvr_block_dma (buff, btr);
  rcvr_spi ();          /* Discard CRC */
  rcvr_spi ();

  return TRUE;          /* Return with success */
}

/*-----------------------------------------------------------------------*/
/* Send a data packet to MMC                                             */
/*-----------------------------------------------------------------------*/
#if _READONLY == 0
static BOOL xmit_datablock (
  const BYTE *buff, /* 512 byte data block to be transmitted */
  BYTE token        /* Data/Stop token */
) {
  BYTE resp, wc;

  if (wait_ready () != 0xFF) return FALSE;

  xmit_spi (token);             /* Xmit data token */
  if (token != 0xFD) {          /* Is data token */
    wc = 0;
    xmit_datablock_dma (buff);
    xmit_spi (0xFF);            /* CRC (Dummy) */
    xmit_spi (0xFF);
    resp = rcvr_spi ();         /* Receive data response */
    if ((resp & 0x1F) != 0x05)  /* If not accepted, return with error */
      return FALSE;
  }

  return TRUE;
}
#endif /* _READONLY */

/*-----------------------------------------------------------------------*/
/* Send a command packet to MMC                                          */
/*-----------------------------------------------------------------------*/
static BYTE send_cmd (
  BYTE cmd,   /* Command byte */
  DWORD arg   /* Argument */
) {
  BYTE n, res;

  if (cmd & 0x80) { /* ACMD<n> is the command sequence of CMD55-CMD<n> */
    cmd &= 0x7F;
    res = send_cmd (CMD55, 0);
    if (res > 1) return res;
  }

  /* Select the card and wait for ready */
  DESELECT ();
  SELECT ();
  if (wait_ready () != 0xFF) return 0xFF;

  /* Send command packet */
  xmit_spi (cmd);                 /* Start + Command index */
  xmit_spi ((BYTE)(arg >> 24));   /* Argument[31..24] */
  xmit_spi ((BYTE)(arg >> 16));   /* Argument[23..16] */
  xmit_spi ((BYTE)(arg >> 8));    /* Argument[15..8] */
  xmit_spi ((BYTE)arg);           /* Argument[7..0] */
  n = 0x01;                       /* Dummy CRC + Stop */
  if (cmd == CMD0) n = 0x95;      /* Valid CRC for CMD0(0) */
  if (cmd == CMD8) n = 0x87;      /* Valid CRC for CMD8(0x1AA) */
  xmit_spi (n);

  /* Receive command response */
  if (cmd == CMD12) rcvr_spi ();  /* Skip a stuff byte when stop reading */
  n = 10;                         /* Wait for a valid response in timeout of 10 attempts */
  do
    res = rcvr_spi ();
  while ((res & 0x80) && --n);

  return res;                     /* Return with the response value */
}

/*--------------------------------------------------------------------------

   Public Functions

---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Initialize Disk Drive                                                 */
/*-----------------------------------------------------------------------*/
DSTATUS disk_initialize (
  BYTE drv  /* Physical drive number (0) */
) {
  BYTE n, cmd, ty, ocr[4];

#if 0
  static BYTE pv;
  pv = get_SOCKWP() | get_SOCKINS();
  if (pv & SOCKINS) /* INS = H (Socket empty) */
    xprintf("not inserted\n");
  else {
    /* INS = L (Card inserted) */
    xprintf("inserted\n");
    if (pv & SOCKWP) /* WP is H (write protected) */
      xprintf("protected\n");
    else
      /* WP is L (write enabled) */
      xprintf("unprotected\n");
  }
#endif

  if (drv) return STA_NOINIT;           /* Supports only single drive */
  if (Stat & STA_NODISK) return Stat;   /* No card in the socket */

  power_on();                           /* Force socket power on */
  for (n = 10; n; n--) rcvr_spi();      /* 80 dummy clocks */

  ty = 0;
  if (send_cmd(CMD0, 0) == 1) {         /* Enter Idle state */
    Timer1 = 100;                       /* Initialization timeout of 1000 msec */
    if (send_cmd(CMD8, 0x1AA) == 1) {   /* SDHC */
      for (n = 0; n < 4; n++) ocr[n] = rcvr_spi();      /* Get trailing return value of R7 resp */
      if (ocr[2] == 0x01 && ocr[3] == 0xAA) {           /* The card can work at vdd range of 2.7-3.6V */
        while (Timer1 && send_cmd(ACMD41, 1UL << 30));  /* Wait for leaving idle state (ACMD41 with HCS bit) */
        if (Timer1 && send_cmd(CMD58, 0) == 0) {        /* Check CCS bit in the OCR */
          for (n = 0; n < 4; n++) ocr[n] = rcvr_spi();
          ty = (ocr[0] & 0x40) ? CT_SD2 | CT_BLOCK : CT_SD2;
        }
      }
    } else {                            /* SDSC or MMC */
      if (send_cmd(ACMD41, 0) <= 1)   {
        ty = CT_SD1; cmd = ACMD41;      /* SDSC */
      } else {
        ty = CT_MMC; cmd = CMD1;        /* MMC */
      }
      while (Timer1 && send_cmd(cmd, 0));         /* Wait for leaving idle state */
      if (!Timer1 || send_cmd(CMD16, 512) != 0)   /* Set R/W block length to 512 */
        ty = 0;
    }
  }
  CardType = ty;
  release_spi();

  if (ty) {                             /* Initialization succeeded */
    Stat &= ~STA_NOINIT;                /* Clear STA_NOINIT */
    AT91_spiSetSpeed(SPI_CSR_NUM, SPI_SCBR_MIN);
  } else {                              /* Initialization failed */
    power_off();
  }

  return Stat;
}

/*-----------------------------------------------------------------------*/
/* Get Disk Status                                                       */
/*-----------------------------------------------------------------------*/
DSTATUS disk_status (
  BYTE drv  /* Physical drive number (0) */
) {
  if (drv) return STA_NOINIT;     /* Supports only single drive */
  return Stat;
}

/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/
DRESULT disk_read (
  BYTE drv,       /* Physical drive number (0) */
  BYTE *buff,     /* Pointer to the data buffer to store read data */
  DWORD sector,   /* Start sector number (LBA) */
  BYTE count      /* Sector count (1..255) */
) {
  if (drv || !count) return RES_PARERR;
  if (Stat & STA_NOINIT) return RES_NOTRDY;

  if (!(CardType & CT_BLOCK)) sector *= 512;  /* Convert to byte address if needed */

  if (count == 1) { /* Single block read */
    if ((send_cmd(CMD17, sector) == 0)  /* READ_SINGLE_BLOCK */
      && rcvr_datablock(buff, 512))
      count = 0;
  } else {          /* Multiple block read */
    if (send_cmd(CMD18, sector) == 0) { /* READ_MULTIPLE_BLOCK */
      do {
        if (!rcvr_datablock(buff, 512)) break;
        buff += 512;
      } while (--count);
      send_cmd(CMD12, 0);        /* STOP_TRANSMISSION */
    }
  }
  release_spi();

  return count ? RES_ERROR : RES_OK;
}

/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/
#if _READONLY == 0
DRESULT disk_write (
  BYTE drv,         /* Physical drive number (0) */
  const BYTE *buff, /* Pointer to the data to be written */
  DWORD sector,     /* Start sector number (LBA) */
  BYTE count        /* Sector count (1..255) */
) {
  if (drv || !count) return RES_PARERR;
  if (Stat & STA_NOINIT) return RES_NOTRDY;
  if (Stat & STA_PROTECT) return RES_WRPRT;

  if (!(CardType & CT_BLOCK)) sector *= 512;  /* Convert to byte address if needed */

  if (count == 1) { /* Single block write */
    if ((send_cmd(CMD24, sector) == 0)        /* WRITE_BLOCK */
      && xmit_datablock(buff, 0xFE))
      count = 0;
  } else {            /* Multiple block write */
    if (CardType & CT_SDC) send_cmd(ACMD23, count);
    if (send_cmd(CMD25, sector) == 0) {       /* WRITE_MULTIPLE_BLOCK */
      do {
        if (!xmit_datablock(buff, 0xFC)) break;
        buff += 512;
      } while (--count);
      if (!xmit_datablock(0, 0xFD))           /* STOP_TRAN token */
        count = 1;
    }
  }
  release_spi();

  return count ? RES_ERROR : RES_OK;
}
#endif /* _READONLY == 0 */

/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/
#if _USE_IOCTL != 0
DRESULT disk_ioctl (
  BYTE drv,		/* Physical drive number (0) */
  BYTE ctrl,		/* Control code */
  void *buff		/* Buffer to send/receive control data */
) {
  DRESULT res;
  BYTE n, csd[16], *ptr = buff;
  WORD csize;

  if (drv) return RES_PARERR;

  res = RES_ERROR;

  if (ctrl == CTRL_POWER) {
    switch (*ptr) {
    case 0:		/* Sub control code == 0 (POWER_OFF) */
      if (chk_power())
        power_off();		/* Power off */
      res = RES_OK;
      break;
    case 1:		/* Sub control code == 1 (POWER_ON) */
      power_on();				/* Power on */
      res = RES_OK;
      break;
    case 2:		/* Sub control code == 2 (POWER_GET) */
      *(ptr+1) = (BYTE)chk_power();
      res = RES_OK;
      break;
    default :
      res = RES_PARERR;
    }
  } else {
    if (Stat & STA_NOINIT) return RES_NOTRDY;

    switch (ctrl) {
    case CTRL_SYNC :		/* Make sure that no pending write process */
      SELECT();
      if (wait_ready() == 0xFF)
        res = RES_OK;
      break;

    case GET_SECTOR_COUNT :	/* Get number of sectors on the disk (DWORD) */
      if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {
        if ((csd[0] >> 6) == 1) {	/* SDC ver 2.00 */
          csize = csd[9] + ((WORD)csd[8] << 8) + 1;
          *(DWORD*)buff = (DWORD)csize << 10;
        } else {					/* SDC ver 1.XX or MMC*/
          n = (csd[5] & 15) + ((csd[10] & 128) >> 7) + ((csd[9] & 3) << 1) + 2;
          csize = (csd[8] >> 6) + ((WORD)csd[7] << 2) + ((WORD)(csd[6] & 3) << 10) + 1;
          *(DWORD*)buff = (DWORD)csize << (n - 9);
        }
        res = RES_OK;
      }
      break;

    case GET_SECTOR_SIZE :	/* Get R/W sector size (WORD) */
      *(WORD*)buff = 512;
      res = RES_OK;
      break;

    case GET_BLOCK_SIZE :	/* Get erase block size in unit of sector (DWORD) */
      if (CardType & CT_SD2) {			/* SDC ver 2.00 */
        if (send_cmd(ACMD13, 0) == 0) {		/* Read SD status */
          rcvr_spi();
          if (rcvr_datablock(csd, 16)) {				/* Read partial block */
            for (n = 64 - 16; n; n--) rcvr_spi();	/* Purge trailing data */
            *(DWORD*)buff = 16UL << (csd[10] >> 4);
            res = RES_OK;
          }
        }
      } else {					/* SDC ver 1.XX or MMC */
        if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {	/* Read CSD */
          if (CardType & CT_SD1) {			/* SDC ver 1.XX */
            *(DWORD*)buff = (((csd[10] & 63) << 1) + ((WORD)(csd[11] & 128) >> 7) + 1) << ((csd[13] >> 6) - 1);
          } else {					/* MMC */
            *(DWORD*)buff = ((WORD)((csd[10] & 124) >> 2) + 1) * (((csd[11] & 3) << 3) + ((csd[11] & 224) >> 5) + 1);
          }
          res = RES_OK;
        }
      }
      break;

    case MMC_GET_TYPE :		/* Get card type flags (1 byte) */
      *ptr = CardType;
      res = RES_OK;
      break;

    case MMC_GET_CSD :		/* Receive CSD as a data block (16 bytes) */
      if (send_cmd(CMD9, 0) == 0		/* READ_CSD */
        && rcvr_datablock(ptr, 16))
        res = RES_OK;
      break;

    case MMC_GET_CID :		/* Receive CID as a data block (16 bytes) */
      if (send_cmd(CMD10, 0) == 0		/* READ_CID */
        && rcvr_datablock(ptr, 16))
        res = RES_OK;
      break;

    case MMC_GET_OCR :		/* Receive OCR as an R3 resp (4 bytes) */
      if (send_cmd(CMD58, 0) == 0) {	/* READ_OCR */
        for (n = 4; n; n--) *ptr++ = rcvr_spi();
        res = RES_OK;
      }
      break;

    case MMC_GET_SDSTAT :	/* Receive SD status as a data block (64 bytes) */
      if (send_cmd(ACMD13, 0) == 0) {	/* SD_STATUS */
        rcvr_spi();
        if (rcvr_datablock(ptr, 64))
          res = RES_OK;
      }
      break;

    default:
      res = RES_PARERR;
    }

    release_spi();
  }

  return res;
}
#endif /* _USE_IOCTL != 0 */

/*-----------------------------------------------------------------------*/
/* Device Timer Interrupt Procedure  (Platform dependent)                */
/*-----------------------------------------------------------------------*/
/* This function must be called in period of 10ms                        */
void disk_timerproc (void) {
  static BYTE pv;
  BYTE n, s;

  n = Timer1;           /* 100Hz decrement timer */
  if (n) Timer1 = --n;
  n = Timer2;
  if (n) Timer2 = --n;

  n = pv;
  pv = get_SOCKWP() | get_SOCKINS();
  // pv = ( ( PIOA.PDSR & SD_SOCKET_WP_PIN ) ? SOCKWP /*protected*/ : 0 /* unprotected */ )
  //	| ( ( PIOA.PDSR & SD_SOCKET_INS_PIN ) ?  SOCKINS /*no card*/ : 0 /* card inserted */ );

  if (n == pv) {          /* Have contacts stabled? */
    s = Stat;

    if (pv & SOCKWP)      /* WP is H (write protected) */
      s |= STA_PROTECT;
    else                  /* WP is L (write enabled) */
      s &= ~STA_PROTECT;

    if (pv & SOCKINS)     /* INS = H (Socket empty) */
      s |= (STA_NODISK | STA_NOINIT);
    else                  /* INS = L (Card inserted) */
      s &= ~STA_NODISK;

    Stat = s;
  }
}


// rtc dummy func
DWORD get_fattime (void) {
  return 0;
}
