//-----------------------------------------------------------
// smpl17(pwm).c
// PA0`PA2PWMMo
//   Copyright (c) BestTechnology CO.,LTD. 2005-2011
//-----------------------------------------------------------
// PWM̃C^[o荞݂ɂĊeXf[eB[XVB
//-----------------------------------------------------------

#if defined (__AT91SAM7X128__) | defined (__AT91SAM7X256__) | defined (__AT91SAM7X512__)
#include <AT91SAM7X.h>
#include <AT91SAM7X_BITDEFINE.h>
#else
#include <AT91SAM7S.h>
#include <AT91SAM7S_BITDEFINE.h>
#endif

#include <piocfg.h>
#include <us.h>
#include <vic.h>

// VAʐM֘A
#define __UDP__
#include "share/us.c"

// PIO[qݒ
#if defined (__AT91SAM7X128__) | defined (y) | defined (__AT91SAM7X512__)
static const TPin pins[] = {
  // PB27 -> PWM0
  {1 << 27, &PIOB, PIOB_PID, PIO_PERIPH_B, PIO_DEFAULT},
  // PB28 -> PWM1
  {1 << 28, &PIOB, PIOB_PID, PIO_PERIPH_B, PIO_DEFAULT},
  // PB29 -> PWM2
  {1 << 29, &PIOB, PIOB_PID, PIO_PERIPH_B, PIO_DEFAULT},
};
#else
static const TPin pins[] = {
  // PA0 -> PWM0
  {1 <<  0, &PIOA, PIOA_PID, PIO_PERIPH_A, PIO_DEFAULT},
  // PA1 -> PWM1
  {1 <<  1, &PIOA, PIOA_PID, PIO_PERIPH_A, PIO_DEFAULT},
  // PA2 -> PWM2
  {1 <<  2, &PIOA, PIOA_PID, PIO_PERIPH_A, PIO_DEFAULT},
};
#endif

int duty[4] = {2,2,2,2};

//========================================================
// MISC...
//========================================================
void wait (int n) {
  volatile int i, j;

  for (i = 0; i < n; i++)
    for (j = 0; j < 1000; j++) ;
}

//========================================================
// PWM
//========================================================
void update_period (int ch, uint16_t period) {
  ch &= 3;
  bool upd = (PWM.CHANNEL[ch].CPRD != period);
  if (period < 1) period = 1;

  if (upd) { while (PWM.SR & (1 << ch)) PWM.DIS = (1 << ch); }
  if (PWM.SR & (1 << ch)) {
    PWM.CHANNEL[ch].CMR |= _PWM_CPD;
    PWM.CHANNEL[ch].CUPD = period;
  } else {
    PWM.CHANNEL[ch].CPRD = period;
  }
  if (upd) PWM.ENA = (1 << ch);
}

void update_duty (int ch, uint16_t dt) {
  ch &= 3;
  if (dt > PWM.CHANNEL[ch].CPRD) dt = PWM.CHANNEL[ch].CPRD;
  if (dt < 2) dt = 2;

  if (PWM.SR & (1 << ch)) {
    PWM.CHANNEL[ch].CMR &= ~_PWM_CPD;
    PWM.CHANNEL[ch].CUPD = dt;
  } else {
    PWM.CHANNEL[ch].CDTY = dt;
  }
}

void __attribute__((section(".ramfunc"))) PWM_irq_handler (void) {
  uint32_t sr = PWM.ISR & PWM.IMR;

  if (sr & _PWM_CHID0) {
    if (duty[0]++ > PWM.CHANNEL[0].CPRD) duty[0] = 0;
    update_duty(0, duty[0]);
  }
  if (sr & _PWM_CHID1) {
    if (duty[1]++ > PWM.CHANNEL[0].CPRD) duty[1] = 0;
    update_duty(1, duty[1]);
  }
  if (sr & _PWM_CHID2) {
    if (duty[2]++ > PWM.CHANNEL[0].CPRD) duty[2] = 0;
    update_duty(2, duty[2]);
  }
}

void PWM_init(void) {
  PMC.PCER = 1 << PWM_PID;

  PWM.DIS = _PWM_CHID0 | _PWM_CHID1 | _PWM_CHID2; // PWM disable

  PWM.CHANNEL[0].CMR  = _PWM_CPRE_MCK1024;
  PWM.CHANNEL[0].CPRD = 300;
  PWM.CHANNEL[0].CDTY = 2;

  PWM.CHANNEL[1].CMR  = _PWM_CPRE_MCK512;
  PWM.CHANNEL[1].CPRD = 300;
  PWM.CHANNEL[1].CDTY = 2;

  PWM.CHANNEL[2].CMR  = _PWM_CPRE_MCK256;
  PWM.CHANNEL[2].CPRD = 300;
  PWM.CHANNEL[2].CDTY = 2;

  AIC.IDCR = 1 << PWM_PID;
  AIC.SVR[PWM_PID] = (void *)PWM_irq_handler;
  AIC.SMR[PWM_PID] = 0;
  AIC.ICCR = 1 << PWM_PID;
  AIC.IECR = 1 << PWM_PID;

  PWM.IER = _PWM_CHID0 | _PWM_CHID1 | _PWM_CHID2; // PWM interrupt
  PWM.ENA = _PWM_CHID0 | _PWM_CHID1 | _PWM_CHID2; // PWM enable
}

//========================================================
// main
//========================================================
void main (void) {
  enableIRQ();
  _us_init (_BAUDRATE, txb, sizeof (txb), rxb, sizeof (rxb));

  // PIO Config
  PIO_Configure (pins, PIO_LISTSIZE (pins));

  // PWM
  PWM_init();

  while(1) {
    if (_us_rx_buff ()) {
      switch (_us_getc ()) {
        case 0x1b:  // I
#if defined (__UDP__)
          udp_disconnect ();
          wait(5000);
#endif
          RSTC.CR = 0xA5000000 | _RSTC_PROCRST | _RSTC_PERRST | _RSTC_EXTRST;
          break;
        case '0':
          if (PWM.SR & _PWM_CHID0) {
            while (PWM.SR & _PWM_CHID0) PWM.DIS = _PWM_CHID0; // ~
          } else {
            PWM.ENA = _PWM_CHID0;                             // Jn
          }
          break;
        case '1':
          if (PWM.SR & _PWM_CHID1) {
            while (PWM.SR & _PWM_CHID1) PWM.DIS = _PWM_CHID1; // ~
          } else {
            PWM.ENA = _PWM_CHID1;                             // Jn
          }
          break;
        case '2':
          if (PWM.SR & _PWM_CHID2) {
            while (PWM.SR & _PWM_CHID2) PWM.DIS = _PWM_CHID2; // ~
          } else {
            PWM.ENA = _PWM_CHID2;                             // Jn
          }
          break;
      }
    }
    _us_printf("PWM duty/period = %4d/%4d, %4d/%4d, %4d/%4d     \r",
      PWM.CHANNEL[0].CDTY, PWM.CHANNEL[0].CPRD,
      PWM.CHANNEL[1].CDTY, PWM.CHANNEL[1].CPRD,
      PWM.CHANNEL[2].CDTY, PWM.CHANNEL[2].CPRD
    );
    wait (100);
  }
}
